using System;
using System.Drawing;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.Orienteering {
  [Item("OrienteeringSolution", "Represents a Orienteering solution which can be visualized in the GUI.")]
  [StorableClass]
  public class OrienteeringSolution : Item {
    public static new Image StaticItemImage {
      get { return HeuristicLab.Common.Resources.VSImageLibrary.Image; }
    }

    [Storable]
    private IntegerVector integerVector;
    public IntegerVector IntegerVector {
      get { return integerVector; }
      set {
        if (integerVector != value) {
          if (integerVector != null) DeregisterIntegerVectorEvents();
          integerVector = value;
          if (integerVector != null) RegisterIntegerVectorEvents();
          OnIntegerVectorChanged();
        }
      }
    }
    [Storable]
    private DoubleMatrix coordinates;
    public DoubleMatrix Coordinates {
      get { return coordinates; }
      set {
        if (coordinates != value) {
          if (coordinates != null) DeregisterCoordinatesEvents();
          coordinates = value;
          if (coordinates != null) RegisterCoordinatesEvents();
          OnCoordinatesChanged();
        }
      }
    }
    [Storable]
    private IntValue startingPoint;
    public IntValue StartingPoint {
      get { return startingPoint; }
      set {
        if (startingPoint != value) {
          if (startingPoint != null) DeregisterStartingPointEvents();
          startingPoint = value;
          if (startingPoint != null) RegisterStartingPointEvents();
          OnStartingPointChanged();
        }
      }
    }
    [Storable]
    private IntValue terminusPoint;
    public IntValue TerminusPoint {
      get { return terminusPoint; }
      set {
        if (terminusPoint != value) {
          if (terminusPoint != null) DeregisterTerminusPointEvents();
          terminusPoint = value;
          if (terminusPoint != null) RegisterTerminusPointEvents();
          OnTerminusPointChanged();
        }
      }
    }
    [Storable]
    private DoubleArray scores;
    public DoubleArray Scores {
      get { return scores; }
      set {
        if (scores != value) {
          if (scores != null) DeregisterScoresEvents();
          scores = value;
          if (scores != null) RegisterScoresEvents();
          OnScoresChanged();
        }
      }
    }
    [Storable]
    private DoubleValue quality;
    public DoubleValue Quality {
      get { return quality; }
      set {
        if (quality != value) {
          if (quality != null) DeregisterQualityEvents();
          quality = value;
          if (quality != null) RegisterQualityEvents();
          OnQualityChanged();
        }
      }
    }

    [StorableConstructor]
    private OrienteeringSolution(bool deserializing) : base(deserializing) { }
    private OrienteeringSolution(OrienteeringSolution original, Cloner cloner)
      : base(original, cloner) {
      this.integerVector = cloner.Clone(original.integerVector);
      this.coordinates = cloner.Clone(original.coordinates);
      this.quality = cloner.Clone(original.quality);
      Initialize();
    }
    public OrienteeringSolution(IntegerVector integerVector, DoubleMatrix coordinates, IntValue startingPoint, IntValue terminusPoint,
      DoubleArray scores, DoubleValue quality)
      : base() {
      this.integerVector = integerVector;
      this.coordinates = coordinates;
      this.startingPoint = startingPoint;
      this.terminusPoint = terminusPoint;
      this.scores = scores;
      this.quality = quality;
      Initialize();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new OrienteeringSolution(this, cloner);
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      Initialize();
    }

    private void Initialize() {
      if (integerVector != null) RegisterIntegerVectorEvents();
      if (coordinates != null) RegisterCoordinatesEvents();
      if (startingPoint != null) RegisterStartingPointEvents();
      if (terminusPoint != null) RegisterTerminusPointEvents();
      if (scores != null) RegisterScoresEvents();
      if (quality != null) RegisterQualityEvents();
    }

    #region Events
    public event EventHandler IntegerVectorChanged;
    private void OnIntegerVectorChanged() {
      var changed = IntegerVectorChanged;
      if (changed != null)
        changed(this, EventArgs.Empty);
    }

    public event EventHandler CoordinatesChanged;
    private void OnCoordinatesChanged() {
      var changed = CoordinatesChanged;
      if (changed != null)
        changed(this, EventArgs.Empty);
    }

    public event EventHandler StartingPointChanged;
    private void OnStartingPointChanged() {
      var changed = StartingPointChanged;
      if (changed != null)
        changed(this, EventArgs.Empty);
    }

    public event EventHandler TerminusPointChanged;
    private void OnTerminusPointChanged() {
      var changed = TerminusPointChanged;
      if (changed != null)
        changed(this, EventArgs.Empty);
    }

    public event EventHandler ScoresChanged;
    private void OnScoresChanged() {
      var changed = ScoresChanged;
      if (changed != null)
        changed(this, EventArgs.Empty);
    }

    public event EventHandler QualityChanged;
    private void OnQualityChanged() {
      var changed = QualityChanged;
      if (changed != null)
        changed(this, EventArgs.Empty);
    }

    private void RegisterIntegerVectorEvents() {
      IntegerVector.ItemChanged += new EventHandler<EventArgs<int>>(IntegerVector_ItemChanged);
      IntegerVector.Reset += new EventHandler(IntegerVector_Reset);
    }
    private void DeregisterIntegerVectorEvents() {
      IntegerVector.ItemChanged -= new EventHandler<EventArgs<int>>(IntegerVector_ItemChanged);
      IntegerVector.Reset -= new EventHandler(IntegerVector_Reset);
    }
    private void RegisterCoordinatesEvents() {
      Coordinates.ItemChanged += new EventHandler<EventArgs<int, int>>(Coordinates_ItemChanged);
      Coordinates.Reset += new EventHandler(Coordinates_Reset);
    }
    private void DeregisterCoordinatesEvents() {
      Coordinates.ItemChanged -= new EventHandler<EventArgs<int, int>>(Coordinates_ItemChanged);
      Coordinates.Reset -= new EventHandler(Coordinates_Reset);
    }
    private void RegisterStartingPointEvents() {
      StartingPoint.ValueChanged += new EventHandler(StartingPoint_ValueChanged);
    }
    private void DeregisterStartingPointEvents() {
      StartingPoint.ValueChanged -= new EventHandler(StartingPoint_ValueChanged);
    }
    private void RegisterTerminusPointEvents() {
      TerminusPoint.ValueChanged += new EventHandler(TerminusPoint_ValueChanged);
    }
    private void DeregisterTerminusPointEvents() {
      TerminusPoint.ValueChanged -= new EventHandler(TerminusPoint_ValueChanged);
    }
    private void RegisterScoresEvents() {
      Scores.ItemChanged += new EventHandler<EventArgs<int>>(Scores_ItemChanged);
      Scores.Reset += new EventHandler(Scores_Reset);
    }
    private void DeregisterScoresEvents() {
      Scores.ItemChanged -= new EventHandler<EventArgs<int>>(Scores_ItemChanged);
      Scores.Reset -= new EventHandler(Scores_Reset);
    }
    private void RegisterQualityEvents() {
      Quality.ValueChanged += new EventHandler(Quality_ValueChanged);
    }
    private void DeregisterQualityEvents() {
      Quality.ValueChanged -= new EventHandler(Quality_ValueChanged);
    }

    private void IntegerVector_ItemChanged(object sender, EventArgs<int> e) {
      OnIntegerVectorChanged();
    }
    private void IntegerVector_Reset(object sender, EventArgs e) {
      OnIntegerVectorChanged();
    }
    private void Coordinates_ItemChanged(object sender, EventArgs<int, int> e) {
      OnCoordinatesChanged();
    }
    private void Coordinates_Reset(object sender, EventArgs e) {
      OnCoordinatesChanged();
    }
    private void StartingPoint_ValueChanged(object sender, EventArgs e) {
      OnStartingPointChanged();
    }
    private void TerminusPoint_ValueChanged(object sender, EventArgs e) {
      OnTerminusPointChanged();
    }
    private void Scores_ItemChanged(object sender, EventArgs<int> e) {
      OnCoordinatesChanged();
    }
    private void Scores_Reset(object sender, EventArgs e) {
      OnCoordinatesChanged();
    }
    private void Quality_ValueChanged(object sender, EventArgs e) {
      OnQualityChanged();
    }
    #endregion
  }
}